local const = require("const")
local json = require("json")
local utils = require("utils3")
local ngx_re_match = ngx.re.match

local function_pattern = [[^(acs:fc:([a-z]{2}-[a-z]+(-\d)?)?:(\d{16})?:)?service:([a-zA-Z_][\w-]{0,127}):function:([a-zA-Z_][\w-]{0,127})$]]
local error_type_unhandled = "Service.Function.Unhandled"
local error_type_handled = "Service.Function.Handled"

local error_mapping = {
    ERROR_FC_REQUEST_NOT_EXIST = const.SERVICE_FUNCTION_NOT_EXIST,
    ERROR_FC_REQUEST_TIMEOUT = const.SERVICE_FUNCTION_TIMEOUT,
}
local error_description_mapping = {
    ERROR_FC_REQUEST_NOT_EXIST = "function does not exist",
    ERROR_FC_REQUEST_TIMEOUT = "function executes timeout",
}

utils.check_request_method("POST")
utils.check_auth()
utils.check_permission("FunctionComputeManage")

local function invoke_function(func)
    local args = {}
    args["serviceName"] = func.service or ""
    args["functionName"] = func.func_name or ""
    args["regionId"] = func.region or ""
    args["accountId"] = func.account_id or ""
    local payload = func.payload or ""
    args["payload"] = ngx.encode_base64(payload)

    local ret, err, cause = utils.dbus_call_edge(
            const.function_compute.wkn,
            const.function_compute.obj,
            const.function_compute.inf,
            "invokeFunction", json.encode(args))
    if not ret then
        return false, err, cause
    end

    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, const.MESSAGE_INVALIDREPLY, "not a JSON object"
    end

    local code = res["code"]
    local data = res["data"]
    if code == 101505 then
        data = ngx.decode_base64(data)
        if not data then
            return false, const.MESSAGE_INVALIDREPLY, "not base64 encoded"
        end
        return false, error_type_handled, data
    end
    if code ~= 0 then
        return false, error_type_unhandled, res["message"]
    end
    if type(data) == "string" then
        data = ngx.decode_base64(data)
        if not data then
            return false, const.MESSAGE_INVALIDREPLY, "not base64 encoded"
        end
    end
    return true, data
end


local function_arn = ngx.var[1]
local items = ngx_re_match(function_arn, function_pattern)
if not items then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "invalid function ARN: " .. function_arn)
end

ngx.req.read_body()
local body = ngx.req.get_body_data()
if body ~= nil and type(body) ~= "string" then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Payload not a JSON")
end

local ok, res, err = invoke_function({
    service = items[5] or "",
    func_name = items[6] or "",
    region = items[2] or "",
    account_id = items[4] or "",
    payload = body,
})

local resp = {}

-- Handle errors.
if not ok then
    if res == error_type_unhandled then
        ngx.header["X-Fc-Error-Type"] = "Unhandled"
        resp["Code"] = error_mapping[err] or err
        resp["Message"] = error_description_mapping[err] or err
        ngx.say(json.encode(resp))
        return
    elseif res == error_type_handled then
        ngx.header["X-Fc-Error-Type"] = "Handled"
        resp["Code"] = ""
        resp["Message"] = ""
        resp["StackTrace"] = ""
        local ok, data = pcall(json.decode, err)
        if ok then
            local err_m = data["errorMessage"]
            resp["Code"] = error_mapping[err_m] or err_m
            resp["Message"] = error_description_mapping[err_m] or err_m
            resp["StackTrace"] = data["stackTrace"]
        end
        ngx.say(json.encode(resp))
        return
    else
        return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, res, err)
    end
end

-- Handle data from the function.
ngx.say(res)
