local json = require("json")
local const = require("const")
local utils = require("utils3")

utils.check_request_method("POST")
utils.check_auth()
utils.check_permission("DeviceManage")

local function list_things(tags)
    local req = { params = { tags = tags } }
    local ret, err, cause = utils.dbus_call_edge(
            const.dimu.wkn, const.dimu.obj, const.dimu.inf,
            "getDeviceState", json.encode(req))
    if not ret then
        return false, err, cause
    end

    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return false, const.MESSAGE_INVALIDREPLY, "not a JSON object"
    end
    if res["code"] ~= 0 then
        local msg = { code = res["code"], message = res["message"] }
        return false, const.MESSAGE_FAILEDREPLY, json.encode(msg)
    end
    if not res["params"] or type(res["params"]) ~= "table" or
            not res["params"]["devices"] or type(res["params"]["devices"]) ~= "table" then
        return false, const.MESSAGE_INVALIDREPLY, "incorrect data format"
    end

    local things = {}
    for _, v in ipairs(res["params"]["devices"]) do
        local product_key = v["productKey"]
        local thing = {}
        thing["ProductName"] = ""
        thing["ProductKey"] = product_key or ""
        thing["DeviceName"] = v["deviceName"] or ""
        thing["NickName"] = v["nickName"] or ""
        thing["DriverId"] = v["driverId"] or ""
        thing["DriverName"] = v["driverName"] or ""
        thing["Tags"] = v["tags"] or {}
        thing["Status"] = "Inactivated"
        thing["ConnectTime"] = ""
        thing["Connected"] = false

        if product_key then
            local ok, tsl_str, cause = utils.get_tsl(product_key)
            if ok then
                local tsl
                ok, tsl = pcall(json.decode, tsl_str)
                if ok then
                    if tsl and tsl["profile"] then
                        thing["ProductName"] = tsl["profile"]["productName"] or ""
                    end
                else
                    utils.log_info("decode tsl %s: %s", tsl_str, tsl)
                end
            else
                utils.log_info("get tsl for %s: %s", product_key, cause)
            end
        end

        if v["localOnline"] == "True" then
            thing["Connected"] = true
        end

        local status
        if v["activation"] == "activated" then
            if v["cloudOnline"] == "True" then
                status = "Online"
            else
                status = "Offline"
            end
        elseif v["activation"] == "activationFailed" then
            status = "Failed"
        end
        if status then
            thing["Status"] = status
        end

        if v["device"][3] then
            thing["ConnectTime"] = v["lastLocalOnlineTime"] or 0
        end
        table.insert(things, thing)
    end
    return true, things, nil
end

ngx.req.read_body()
local body = ngx.req.get_body_data()
if not body then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "no body found")
end

local ok, args = pcall(json.decode, body)
if not ok then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "not a JSON body")
end

local tags = args["Tags"]
if type(tags) ~= "table" then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Tags not a object or is empty")
end
local k = next(tags)
if not k then
    return utils.error_exit(
            ngx.HTTP_BAD_REQUEST,
            const.INVALID_PARAMETER,
            "Tags not a object or is empty")
end

local ok, things, err = list_things(tags)
if not ok then
    return utils.error_exit(ngx.HTTP_INTERNAL_SERVER_ERROR, things, err)
end

local resp = {}
resp["Data"] = {}
resp["Data"]["Things"] = things
local str = json.encode(resp)
ngx.header['Content-Type'] = 'application/json; charset=utf-8'
ngx.say(str)