--- List devices in some driver that support upgrade.

local const = require("const")
local json = require("json")
local utils = require("utils")
local request_method = ngx.var.request_method

utils.check_login()
utils.check_permission("OTAManage")

if request_method ~= "POST" then
    return utils.error_exit(405, "unsupported request method: " .. request_method)
end

local utils_header = require("utils_header")
local params = utils_header.check_public_body_data()
local driver_id = params["DriverId"]
if type(driver_id) ~= "string" or #driver_id == 0 then
    return utils.error_exit(400, "DriverId is not string or is empty")
end

local function list_ota_devices(driver_id)
    local params = string.format('{"driverId":"%s"}', driver_id)
    local ok, ret, err = utils.dbus_call_edge(
            const.device_ota.wkn,
            const.device_ota.obj,
            const.device_ota.inf,
            const.device_ota.method.list_devices,
            params)
    if not ok then
        return nil, err
    end
    local ok, res = pcall(json.decode, ret[1])
    if not ok or type(res) ~= "table" then
        return nil, "response not a JSON object"
    end
    if res["code"] ~= 0 then
        return nil, tostring(res["message"])
    end
    local data = res["data"]
    if not data or type(data) ~= "table" then
        return nil, "data in response not a array"
    end
    return res["data"]
end

local drivers, err = list_ota_devices(driver_id)
if drivers == nil then
    return utils.error_exit(500, err)
end

local resp = {
    code = 200,
    message = "success",
    data = { List = {} }
}
for _, v in ipairs(drivers) do
    local item = {
        ProductKey = v["productKey"],
        DeviceName = v["deviceName"],
        NickName = v["nickName"],
        DriverId = v["driverId"],
        DriverName = v["driverName"],
        Tags = v["tags"],
        ConnectTime = v["lastLocalOnlineTime"],
        FirmwareVersion = v["version"],
        UpgradeProgress = v["progress"],
        LastUpgradeTime = v["otaTime"] or "",
    }

    local status
    if v["localOnline"] == "True" then
        status = "Online"
    else
        status = "Offline"
    end
    item["Status"] = status

    local connected
    if v["cloudOnline"] == "True" then
        connected = true
    else
        connected = false
    end
    item["Connected"] = connected

    table.insert(resp.data.List, item)
end
ngx.say(json.encode(resp))
