--- For Open APIs 2020-04-30.
-- Utils 0.3.0 is only used in Open APIs 2020-04-30.

local const = require("const")
local json = require("json")
local ldbus = require("ldbus")
-- Includes utils2 to offer those common APIs.
local utils2 = require("utils2")

local format = string.format

local utils = { _version = "0.3.0" }

-- Common APIs
function utils.is_array_table(t)
    return utils2.is_array_table(t)
end

function utils.is_in_array(array, item)
    return utils2.is_in_array(array, item)
end

function utils.split_str(str, sep)
    return utils2.split_str(str, sep)
end

function utils.hmac_sha256(key, data)
    return utils2.hmac_sha256(key, data)
end

function utils.log_debug(fmt, ...)
    utils2.log_debug(fmt, ...)
end

function utils.log_info(fmt, ...)
    utils2.log_info(fmt, ...)
end

function utils.log_warn(fmt, ...)
    utils2.log_warn(fmt, ...)
end

function utils.log_error(fmt, ...)
    utils2.log_error(fmt, ...)
end

function utils.log_critical(fmt, ...)
    utils2.log_critical(fmt, ...)
end

-- V3 APIs
function utils.error_exit(status, code, message)
    local res = {}
    res["Code"] = code
    res["Message"] = message
    ngx.status = status
    ngx.header['Content-Type'] = 'application/json; charset=utf-8'
    ngx.say(json.encode(res))
    -- https://openresty-reference.readthedocs.io/en/latest/Lua_Nginx_API/#ngxexit
    ngx.exit(ngx.HTTP_OK)
end

function utils.check_request_method(expect)
    local method = ngx.var.request_method
    if method ~= expect then
        return utils.error_exit(
                ngx.HTTP_NOT_ALLOWED,
                const.METHOD_NOT_ALLOWED,
                "not allowed HTTP method: " .. method)
    end
end

function utils.check_auth()
    local resty_cookie = require("resty.cookie")
    local cookie, err = resty_cookie:new()
    if not cookie then
        return utils.error_exit(
                ngx.HTTP_INTERNAL_SERVER_ERROR,
                const.COOKIE_HANDLER_CREATE,
                err)
    end

    print("client ip is: " .. ngx.var.remote_addr)
    -- get access cookie
    local token = cookie:get("token")
    if not token then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "unauthorized")
    end

    local memkv = ngx.shared.memkv
    local perm = memkv:get(token)
    if not perm then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "unauthorized")
    end
    return true, token, perm
end

function utils.check_permission(...)
    local ok, token, perm = utils.check_auth()
    if not ok then
        return utils.error_exit(
                ngx.HTTP_UNAUTHORIZED,
                const.UNAUTHORIZED,
                "unauthorized")
    end

    if not perm then
        return utils.error_exit(
                ngx.HTTP_FORBIDDEN,
                const.FORBIDDEN,
                "permission denied")
    end

    local perm_list = utils.split_str(perm, ",")
    for _, item in ipairs { ... } do
        if (utils.is_in_array(perm_list, item)) then
            return true, token
        end
    end
    return utils.error_exit(
            ngx.HTTP_FORBIDDEN,
            const.FORBIDDEN,
            "permission denied")
end

local function get_default_dbus_connection()
    if default_dbus_connection then
        return default_dbus_connection, nil
    end

    local bus = "unix:path=/tmp/var/run/mbusd/mbusd_socket"
    local conn, err_c = ldbus.connection.open(bus)
    if not conn then
        return nil, format("open connection to %s: %s", bus, tostring(err_c))
    end

    local reg, err_r = ldbus.bus.register(conn)
    if not reg then
        return nil, "register connection: " .. tostring(err_r)
    end

    default_dbus_connection = conn
    return conn, nil
end

function utils.dbus_call_edge(service, path, iface, method, ...)
    local failure = function(code, cause)
        utils.log_error(code .. ": " .. cause)
        return nil, code, cause
    end

    local conn, err_c = get_default_dbus_connection()
    if not conn then
        return failure(const.CONNECTION_CREATE, err_c)
    end

    local msg = ldbus.message.new_method_call(service, path, iface, method)
    if not msg then
        local err_m = format("cannot create method call for (%s, %s, %s.%s)",
                service, path, iface, method)
        return failure(const.MESSAGE_CREATE, err_m)
    end

    local iter = ldbus.message.iter.new()
    msg:iter_init_append(iter)

    for _, arg in ipairs { ... } do
        iter:append_basic(arg)
    end

    local timeout = 20
    if (service == const.function_compute.wkn) then
        timeout = 1800
    end

    local reply, err_r = conn:send_with_reply_and_block(msg, timeout)
    if not reply then
        local service = utils2.is_service_not_exist_err(err_r)
        if service then
            err_r = format("upstream service %s not exist", service)
            return failure(const.SERVICE_NOT_EXIST, err_r)
        end
        return failure(const.MESSAGE_SENDANDREPLY, err_r)
    end

    if not reply:iter_init(iter) then
        return failure(const.MESSAGE_INVALIDREPLY, "no arguments found in reply")
    end

    local ret = {}
    while iter:get_arg_type() ~= nil do
        table.insert(ret, iter:get_basic())
        iter:next()
    end
    return ret
end

function utils.set_data(key, val)
    local ret, _, cause = utils.dbus_call_edge(
            "iot.dmp.configmanager",
            "/iot/dmp/configmanager",
            "iot.dmp.configmanager",
            "set_config", key, val)

    if not ret then
        return false, const.SERVICE_DATA_SET, format("failed to set data: %s", cause)
    end

    local code = tonumber(ret[1])
    if code ~= 0 then
        return false, const.SERVICE_DATA_SET, format("failed to set data: errno=%s", code)
    end

    return true
end

function utils.get_data(key)
    local ret, _, cause = utils.dbus_call_edge(
            "iot.dmp.configmanager",
            "/iot/dmp/configmanager",
            "iot.dmp.configmanager",
            "get_config",
            key)
    if not ret then
        return false, const.SERVICE_DATA_GET, format("failed to get data: %s", cause)
    end

    local code = tonumber(ret[1])
    if code == 100100 then
        return false, const.SERVICE_DATA_KEY_NOT_EXIST, "key does not exit"
    end
    if code ~= 0 then
        return false, const.SERVICE_DATA_GET, format("failed to get data: errno=%s", code)
    end
    return true, ret[2]
end

local function is_match_str(str, pattern, len_min, len_max)
    if not str or type(str) ~= "string" then
        return false
    end

    local len = string.len(str)
    if len < len_min or len > len_max then
        return false
    end
    return string.match(str, pattern)
end

function utils.is_valid_pk(pk)
    return is_match_str(pk, "^%w+$", 11, 11)
end

function utils.is_valid_dn(dn)
    return is_match_str(dn, "^[%w%_%-%.%@%:]+$", 4, 32)
end

function utils.is_valid_ds(ds)
    return is_match_str(ds, "^%w+$", 32, 32)
end

function utils.is_valid_username(username)
    return is_match_str(username, "^%w+$", 32, 32)
end

function utils.is_valid_password(password)
    -- 64 chars password
    if type(password) == "string" and string.len(password) == 64 then
        return true
    end
    return is_match_str(password, "^%w+$", 5, 20)
end

function utils.get_tsl(product)
    return utils.get_data(const.tsl.prefix .. product)
end

return utils