#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import time
import threading
import json
import paho.mqtt.client as mqtt
from runtime_utils import utils
from runtime_utils import sync
from runtime_utils import thread
from runtime_utils import config
from runtime_utils import fc_error as exception


class MqttWrapper(object):
    _instance = None

    def __del__(self):
        self.mqtt_client.disconnect()

    def __new__(self, *args, **kw):
        if not self._instance:
            self._instance = super(MqttWrapper, self).__new__(self)
        return self._instance

    def __init__(self, ip_address, port, service_cb, driver_id=None, driver_name=None):
        if driver_id is None:
            self.driver_id = os.environ.get("FUNCTION_ID")
        else:
            self.driver_id = driver_id
        if driver_name is None:
            self.driver_name = os.environ.get("FUNCTION_NAME")
        else:
            self.driver_name = driver_name
        self.requestLock = threading.Lock()
        self.message_id = 0
        self.request_sync = {}
        self.threadPool = thread.ThreadPool(config.THREAD_POOL_THREAD_NUM, config.THREAD_POOL_QUEUE_NUM)
        self.username = "driver&" + self.driver_name + "&" + self.driver_id
        self.password = self.driver_id
        self.service_cb = service_cb
        self.mqtt_client = mqtt.Client(self.driver_id)
        self.mqtt_client.on_message = self.on_message
        self.mqtt_client.on_connect = self.on_connect
        self.mqtt_client.on_disconnect = self.on_disconnect
        self.mqtt_client.on_publish = self.on_publish
        self.mqtt_client.on_subscribe = self.on_subscribe
        self.mqtt_client.username_pw_set(self.username, self.password)
        self.t = threading.Thread(target=self.mqtt_loop, args=())
        self.t.setDaemon(True)
        self.t.start()
        self.inited = False
        self.requestLock.acquire()
        try:
            self.mqtt_client.connect(ip_address, port, 120)
        except Exception as e:
            logging.error(e)

    def mqtt_loop(self):
        self.mqtt_client.loop_forever()

    def get_topic_type(self, topic):
        if (topic.startswith("edge/iot/driver/proxy/x/response/driver/" + self.driver_id) or
                topic.startswith("edge/iot/driver/proxy/x/response/device/" + self.driver_id)):
            return 1
        elif topic.startswith("edge/iot/driver/proxy/x/request/device/" + self.driver_id):
            return 2
        else:
            return 0

    def get_down_link_params(self, topic):
        topic_list = topic.split("/")
        length = len(topic_list)
        if length > 13 and topic_list[10] == "callservice":
            if topic_list[13] == "property" and length > 14:
                return topic_list[8], topic_list[9], topic_list[14]
            else:
                return topic_list[8], topic_list[9], topic_list[13]

    def on_message(self, mqtt_client, obj, msg):
        logging.debug(msg.topic + " " + str(msg.qos) + " " + str(msg.payload))
        topic_type = self.get_topic_type(msg.topic)

        if topic_type == 1:
            message = json.loads(msg.payload.decode("utf-8"))
            if "data" in message:
                data = message["data"]
            else:
                data = None
            message_id = message["id"]
            try:
                request_sync = self.request_sync[message_id]
                request_sync.set(data)
            except Exception as e:
                logging.info("receive message, can't found message_id. %s", e)
        elif topic_type == 2:
            product_key, device_name, method_name = self.get_down_link_params(msg.topic)
            logging.debug(product_key+" "+device_name+" "+method_name)
            self.threadPool.addTask(self.service_cb,
                                    product_key=product_key,
                                    device_name=device_name,
                                    topic=msg.topic,
                                    method_name=method_name,
                                    params=msg.payload.decode("utf-8"))

    def on_connect(self, mqtt_client, obj, flags, rc):
        logging.info("connected, return rc: " + str(rc))
        topic = "edge/iot/driver/proxy/x/+/+/"+self.driver_id+"/+/+/#"
        mqtt_client.subscribe(topic, 0)
        if self.inited is False:
            self.inited = True
            self.requestLock.release()

    def on_disconnect(self, mqtt_client, obj, rc):
        logging.debug("disconnect rc: " + str(rc))
        os._exit(0)

    def on_publish(self, obj, mid):
        logging.debug("publish mid: " + str(mid))

    def on_subscribe(self, mqtt_client, obj, mid, granted_qos):
        logging.debug("Subscribed: " + str(mid) + " " + str(granted_qos))

    def send_message(self, topic, data=None):
        with self.requestLock:
            if "id" not in data:
                message_id = str(self.message_id)
                data["id"] = message_id
                self.message_id += 1
            data_str = json.dumps(data)
            logging.debug("request async: " + topic + " " + data_str)
            self.mqtt_client.publish(topic, data_str, 0)

    def send_message_sync(self, topic, data=None):
        with self.requestLock:
            message_id = str(self.message_id)
            if data is None:
                data = {
                    "id": message_id,
                    "version": "1.0",
                    "params": {
                    },
                }
            else:
                data["id"] = message_id
            self.message_id += 1
            request_sync = sync.SyncMsg_Event()
            request_sync.clear()
            self.request_sync[message_id] = request_sync
            logging.debug("request sync: " + topic + " " + json.dumps(data))
            self.mqtt_client.publish(topic, json.dumps(data), 0)
            sync_msg = request_sync.wait(config.METHOD_SYNC_TIMEOUT)
            self.request_sync.pop(message_id)

        return sync_msg

    def get_config(self):
        topic = "edge/driver/" + self.driver_id + "/x/x/request/iot/driver/proxy/x/get_config"
        data = self.send_message_sync(topic)
        return json.dumps(data)

    def get_tsl_info(self, product_key):
        topic = "edge/driver/" + self.driver_id + "/" + product_key + "/x/request/iot/driver/proxy/x/get_tsl"
        data = self.send_message_sync(topic)
        return json.dumps(data)

    def get_tsl_ext_info(self, product_key):
        topic = "edge/driver/" + self.driver_id + "/" + product_key + "/x/request/iot/driver/proxy/x/get_tsl_ext"
        data = self.send_message_sync(topic)
        return json.dumps(data)

    def online(self, product_key, device_name):
        topic = "edge/device/" + self.driver_id + "/" + product_key + "/" + device_name + "/request/iot/driver/proxy/x/online/"
        data = self.send_message_sync(topic)
        return data

    def offline(self, product_key, device_name):
        topic = "edge/device/" + self.driver_id + "/" + product_key + "/" + device_name + "/request/iot/driver/proxy/x/offline"
        data = self.send_message_sync(topic)
        return data

    def reportProperties(self, product_key, device_name, properties, timestamp=None):
        if isinstance(properties, dict) is False or len(properties) is 0:
            logging.error(
                "device(%s:%s):reportProperties,params type is invalid, should be dict or not empty" % (product_key, device_name))
            return None
        topic = "edge/device/"+ self.driver_id + "/" + product_key + "/"\
                + device_name + "/broadcast/thing/" + product_key + "/"\
                + device_name + "/property"
        tmpPropertiesDict = {}
        if (timestamp is not None) and (isinstance(timestamp, int) is True):
            tmp_time = timestamp
        else:
            tmp_time = int(round(time.time() * 1000))
        for key, value in properties.items():
            tmpPropertiesDict[key] = {}
            tmpPropertiesDict[key]['value'] = value
            tmpPropertiesDict[key]['time'] = tmp_time

        prop_data = {
            "version": "1.0",
            "params": tmpPropertiesDict
        }
        data = self.send_message(topic, data=prop_data)
        # return data

    def reportEvent(self, product_key, device_name, eventName, eventDict, timestamp=None):
        if (isinstance(eventDict, dict) is False) or (isinstance(eventName, str) is False):
            logging.error(
                "device(%s:%s):reportEvent,params type is invalid, eventName should be str and event should be dict"
                % (product_key, device_name))
            return None
        topic = "edge/device/" + self.driver_id + "/" + product_key + "/"\
                + device_name + "/broadcast/thing/" + product_key + "/"\
                + device_name + "/event/" + eventName
        if (timestamp is not None) and (isinstance(timestamp, int) is True):
            tmp_time = timestamp
        else:
            tmp_time = int(round(time.time() * 1000))
        event_data = {
            "version": "1.0",
            "params": {
                'value': eventDict,
                'time': tmp_time
            }
        }
        self.send_message(topic, data=event_data)


class LedaSubDevice(object):
    def __init__(self, mqtt_client, device_name, product_key, callback_object):
        self.mqtt_client = mqtt_client
        self.product_key = product_key
        self.device_name = device_name
        self.callback_funs = callback_object
        self.onofflineLock = threading.Lock()

    def device_connect(self):
        with self.onofflineLock:
            if self.mqtt_client is not None:
                self.mqtt_client.online(self.product_key, self.device_name)

    def device_disconnect(self):
        with self.onofflineLock:
            if self.mqtt_client is not None:
                self.mqtt_client.offline(self.product_key, self.device_name)

    def online(self):
        self.device_connect()

    def offline(self):
        self.device_disconnect()

    def reportProperties(self, propertiesDict, timestamp=None):
        if self.mqtt_client is not None:
            self.mqtt_client.reportProperties(self.product_key, self.device_name, propertiesDict, timestamp)

    def reportEvent(self, eventName, eventDict, timestamp=None):
        if self.mqtt_client is not None:
            self.mqtt_client.reportEvent(self.product_key, self.device_name, eventName, eventDict, timestamp)

    def callServices(self, methodName, inMsg):
        codeInfoDict = {
                exception.LE_SUCCESS: 'successfully',
                exception.LE_ERROR_INVAILD_PARAM: 'invalid params',
                exception.LE_ERROR_UNKNOWN: 'exec failed',
                exception.LE_ERROR_TIMEOUT: 'timeout',
                exception.LEDA_ERROR_PROPERTY_NOT_EXIST: 'property not exist',
                exception.LEDA_ERROR_PROPERTY_READ_ONLY: 'property read only',
                exception.LEDA_ERROR_PROPERTY_WRITE_ONLY: 'property write only',
                exception.LEDA_ERROR_SERVICE_NOT_EXIST: 'service not exist',
                exception.LEDA_ERROR_SERVICE_INPUT_PARAM: 'invalid service input params'
            }
        try:
            inMsg_json = json.loads(inMsg)
            inArgs = inMsg_json["params"]
            if "id" in inMsg_json:
                id = inMsg_json["id"]
            else:
                id = 0

            if methodName == "get":
                code, retInfo = self.callback_funs.getProperties(inArgs)
            elif methodName == "set":
                code, retInfo = self.callback_funs.setProperties(inArgs)
                retInfo = {}
            else:
                code, retInfo = self.callback_funs.callService(methodName, inArgs)

            if code in codeInfoDict:
                message = codeInfoDict[code]
            else:
                if ((code >= exception.LEDA_ERROR_USER_MIN) and (code <= exception.LEDA_ERROR_USER_MAX)):
                    message = ''
                else:
                    code = exception.LEDA_ERROR_UNKNOWN
                    message = codeInfoDict[code]
            retDict = {
                "id": id,
                "code": code,
                "message": message,
                "data": retInfo
            }
            return json.dumps(retDict, cls=utils.Json_Encoder, ensure_ascii=False)
        except Exception as e:
            retDict = {
                "id": id,
                "code": exception.LE_ERROR_UNKNOWN,
                "message": str(e),
                "data": {}
            }
            return json.dumps(retDict, ensure_ascii=False)


class LedaModule(object):
    def __init__(self):
        self.mqtt_client = None
        self.deviceServiceDictLock = threading.Lock()
        self.device_service_dict = {}

    def moduleInit(self, driver_id=None, driver_name=None):
        if "FCBASE_IPADDR" in os.environ:
            ipaddr = os.environ.get("FCBASE_IPADDR")
        else:
            ipaddr = config.DEFAULT_IP_ADDRESS
        self.mqtt_client = MqttWrapper(ipaddr, config.DEFAULT_PORT, self.service_cb, driver_id, driver_name)

    def getConfig(self):
        if self.mqtt_client is not None:
            return self.mqtt_client.get_config()

    def getTSL(self, productKey):
        if self.mqtt_client is not None:
            return self.mqtt_client.get_tsl_info(productKey)

    def getTSLConfig(self, productKey):
        if self.mqtt_client is not None:
            return self.mqtt_client.get_tsl_ext_info(productKey)

    def deviceRegister(self, deviceName, productKey, profile, deviceCallBack):
        if (isinstance(deviceName, str) is False) or (isinstance(productKey, str) is False) or (isinstance(deviceCallBack, BaseDeviceCallback) is False):
            logging.error("deviceRegister: input args type is invalid")
            return None
        with self.deviceServiceDictLock:
            pk_dn = productKey + deviceName
            if pk_dn in self.device_service_dict:
                sudDevice = self.device_service_dict[pk_dn]
                sudDevice.online()
            else:
                sudDevice = LedaSubDevice(self.mqtt_client, deviceName, productKey, deviceCallBack)
                sudDevice.online()
                self.device_service_dict[pk_dn] = sudDevice

            return sudDevice

    def deviceUnregister(self, subDevice):
        with self.deviceServiceDictLock:
            pk_dn = subDevice.product_key + subDevice.device_name
            if pk_dn in self.device_service_dict:
                del self.device_service_dict[pk_dn]
            subDevice.offline()

    def service_cb(self, product_key, device_name, topic, method_name, params):
        pk_dn = product_key + device_name
        if pk_dn in self.device_service_dict:
            sudDevice = self.device_service_dict[pk_dn]
            reply_data = sudDevice.callServices(method_name, params)
            topic_list = topic.split("/")
            src = topic_list[1] + "/" +  topic_list[2] + "/" + topic_list[3] + "/" + topic_list[4]
            dest = topic_list[6] + "/" +  topic_list[7] + "/" + topic_list[8] + "/" + topic_list[9]
            length = len(topic_list)
            if length > 14:
                topic_new = "edge/" + dest + "/response/" + src + "/callservice/thing/service/property/" + method_name
            else:
                topic_new = "edge/" + dest + "/response/" + src + "/callservice/thing/service/" + method_name
            if self.mqtt_client is not None:
                self.mqtt_client.send_message(topic_new, json.loads(reply_data))


class BaseDeviceCallback(object):
    pass
